/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#define fesslix_fesslix_CPP

#include "fesslix.h"
#include "flxversion.h"
#include "flxobjrbrv.h"
#include "flxobjmtx.h"
#include "flxdefault.h"
#include "flxobjects.h"
#include "flxstringfun_fun.h"
#include "flxostools.h"
#include "flxMtx_Eigen.h"

#include <fstream>
#include <limits>
#include <iostream>


#ifdef __unix__
  #include <dlfcn.h>                // dlopen functionality
  void* handle = NULL;
#elif defined(__WINDOWS__)
  #include <windows.h>
  HMODULE sdl_library = NULL;
#endif


#if FLX_BOOST_PO
  #include <boost/program_options.hpp>
  namespace po = boost::program_options;
#endif
  
using namespace std;

int FesslixMain::Cinst = 0;
bool fesslix_logOutput=false;        // true, if output is logged and log-stream is not cout
bool fesslix_useScreen=true;        // true, cout is std::cout (otherwise cout is redirected to the log-file)
std::ostream* flxcout = NULL;
int flx_engine_init_count = 0;        
FLXKERNEL_EXPORT bool fesslix_pauseAtEnd=DEFAULT_PAUSE_AT_END;
FLXKERNEL_EXPORT std::string fesslix_CMDstr;
FLXKERNEL_EXPORT std::vector<std::string> fesslix_fileVec;
FLXKERNEL_EXPORT bool fesslix_call_interact = false;
FLXKERNEL_EXPORT bool FlxVi_Deactivate = false;
FesslixMain* FlxObjInteractive::fesslix_ptr = NULL;

/**
* @brief execute this command at the very beginning
*/
void flxengine_load();
/**
* @brief opens the log-file and outputs the header of the log
*/
void flx_create_log(const std::string& logFile, const bool logTrunc);
/**
* @brief returns the option from the argument array
* @param argc number of arguments in the array
* @param argv array with arguments
* @param pos position of the occurence of CMD
* @param value string with the value of the CMD
* @return false if there was no CMD in the command line
*/
bool flx_arg_CMD(int argc, char** argv, int& pos, std::string& valueS );

#if FLX_BOOST_PO
  void flx_print_help(po::options_description &visible);
#else 
  void flx_print_help();
#endif
  
  
  
  
FesslixMain::FesslixMain () 
{
  // check for multiple instances
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxData' created ...";
      throw FlxException("FesslixMain::FesslixMain", ssV.str() );
    }
    
  // create readers ...
    funReader = new FlxFunctionReader();
      FlxReadManager::set_funReader(funReader);
  FlxObjReadLoadLib::AddOnManager = &AddOnManager;

  // create the evaluation class
    EvaluateCmd = new FlxEvaluateCmd();

  // create ObjReaders
    // The LoadLib-Objects
      EvaluateCmd->get_ObjReadBox()->insert("loadlib", new FlxObjReadLoadLib(EvaluateCmd->get_ObjReadBox()));
      EvaluateCmd->get_ObjReadBox()->insert("loadlibfun", new FlxObjReadLoadLibFun(EvaluateCmd->get_ObjReadBox()));
    // Common Objects
      load_FlxReaders( new FlxCreateObjReaders_Common() );
    // Mtx Objects
      load_FlxReaders( new FlxCreateObjReaders_Mtx() );
    // RBRV Objects
      load_FlxReaders( new FlxCreateObjReaders_RBRV() );
    // FlxString Objects
      load_FlxReaders( new FlxCreateObjReaders_FlxString() );
    // The interactive-Object
      EvaluateCmd->get_ObjReadBox()->insert("interactive", new FlxObjReadInteractive());
    
  FlxObjInteractive::set_fesslix_ptr(this);
}

void FesslixMain::initialize(const ostreamp coutV, const ostreamp cerrV, const string& gaussFile)
{
  // check for multiple runs
    static int Cruns = 0;
    ++Cruns;
    if (Cruns > 1) {
      std::ostringstream ssV;
      ssV << "FesslixMain::initialize can only be run once ...";
      throw FlxException("FesslixMain::initialize", ssV.str() );
    }
  // initialise output-streams
    dataBox.OstreamBox.insert("cout", coutV);
    dataBox.OstreamBox.insert("cerr", cerrV);
    dataBox.OstreamBox.insert("dummy",new flxDummyOstream());
    dataBox.OstreamBox.insert("log", &(GlobalVar.slog(0)) );
  // Read Gauss-Point Information
    dataBox.GaussInt.ReadGP(0,gaussFile);  
}

void FesslixMain::load_FlxReaders(FlxCreateObjReaders* cor)
{
  cor->createFunReaders( &dataBox);    
  cor->createObjReaders( EvaluateCmd->get_ObjReadBox() );
  delete cor; 
}

FesslixMain::~FesslixMain () {
  AddOnManager.delete_AddOns();        // free the loaded AddOns
  delete EvaluateCmd;
  delete funReader;
}

void FesslixMain::evaluate () {
  FlxObjBase* ob=NULL;
  if (reader == NULL) {        // Check if a reader is available
    ostringstream ssV_2;
    ssV_2 << "No reader available.";
    throw FlxException("FesslixMain::evaluate_1", ssV_2.str() ); 
  }
  try {
    while ( reader->getNextType() != ReadStream::ENDOFFILE ) {
      // print the prompt for output of input to log
        GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
      ob = EvaluateCmd->evaluateCmd();
      ob->exec();
      GlobalVar.logLevel_strong_reset();
      GlobalVar.slog_flush();
      delete ob; ob=NULL;
    }
  } catch ( FlxEndE &e) {
    FLXMSG("FesslixMain::evaluate_2",1);
    GlobalVar.logLevel_strong_reset();
    if (ob) delete ob;
    throw;
  } catch ( FlxExitE &e) {
    FLXMSG("FesslixMain::evaluate_3",1);
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "exit: 'exit;' was called" << std::endl;
    if (ob) delete ob;
  } catch ( FlxException &e ) {
    FLXMSG("FesslixMain::evaluate_4",1);
    GlobalVar.logLevel_strong_reset();
    if (ob) delete ob;
    throw;
  }
}

void FesslixMain::evaluate(ReadStream* readerV)
{
  dataBox.ReadManager.push(readerV);
  try {
    evaluate();
    dataBox.ReadManager.pop();
  } catch ( FlxEndE &e) {
    FLXMSG("FesslixMain::evaluate_101",1);
    dataBox.ReadManager.pop();
    throw;
  } catch ( FlxException &e ) {
    FLXMSG("FesslixMain::evaluate_102",1);
    dataBox.ReadManager.pop();
    throw;
  }
}

FlxObjBase* FesslixMain::read_objects(ReadStream* readerV)
{
  dataBox.ReadManager.push(readerV);
  FlxCodeBlock* obR = new FlxCodeBlock(true);
  try {
    if (reader == NULL) {        // Check if a reader is available
      ostringstream ssV_2;
      ssV_2 << "No reader available.";
      throw FlxException("FesslixMain::read_objects_1", ssV_2.str() ); 
    }
    while ( reader->getNextType() != ReadStream::ENDOFFILE ) {
      FlxObjBase* ob = EvaluateCmd->evaluateCmd();
      obR->attach_obj(ob);
    }
    dataBox.ReadManager.pop();
    return obR;
  } catch ( FlxException &e) {
    FLXMSG("FesslixMain::read_objects_101",1);
    delete obR;
    dataBox.ReadManager.pop();
    throw;
  }
}

void FesslixMain::evaluate_expr(const string exprStr)
{
  ReadStream Freader(exprStr);
  dataBox.ReadManager.push(&Freader);
  FlxObjBase* ob=NULL;
  try {
    while ( Freader.getNextType() != ReadStream::ENDOFFILE ) {
      // print the prompt for output of input to log
        GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
      ob = EvaluateCmd->evaluateCmd();
      ob->exec();
      delete ob; ob=NULL;
    }
  } catch ( FlxExitE &e) {
    FLXMSG("FesslixMain::evaluate_expr_1",1);
    if (ob) delete ob;
  } catch ( FlxEndE &e) {
    FLXMSG("FesslixMain::evaluate_expr_2",1);
    if (ob) delete ob;
    dataBox.ReadManager.pop();
    throw;
  } catch (...) {
    if (ob) delete ob;
    dataBox.ReadManager.pop();
    throw;
  };
  dataBox.ReadManager.pop();
}


#if FLX_DEBUG
void FesslixMain::test()
{  
  GlobalVar.slog(5) << "Fesslix: test-section ------------ START -----------------------" << std::endl;
  
  
  GlobalVar.slog(5) << "Fesslix: test-section ------------ END -------------------------" << std::endl;
}
#endif


void flx_print_author()
{
  *(GlobalVar.get_cout()) << "Fesslix was written by" << std::endl << 
   "    Wolfgang Betz <w.betz@fesslix.org>" << std::endl <<
   "Please use https://bugs.launchpad.net/fesslix to report bugs." << std::endl
   ;
}

#if FLX_BOOST_PO
  void flx_print_help(po::options_description &visible)
  {
    std::string ts;
    ts = "fesslix";
    *(GlobalVar.get_cout()) << "Usage: " << ts << " [options] files [--CMD code]" << std::endl << std::endl
      << "Fesslix - Stochastic Analysis" << std::endl << std::endl
      << visible << std::endl
      ;
  }
#else
  void flx_print_help()
  {
    std::string ts;
    ts = "fesslix";
    *(GlobalVar.get_cout()) << "Usage: " << ts << " files" << std::endl << std::endl
      << "Fesslix - Stochastic Analysis" << std::endl << std::endl
      ;
  }
#endif

void flx_logInfo(ostream& lout, bool pversion)
{
  lout << " Fesslix:" << std::endl;
  lout << "   compiled with the options ..." << std::endl;
  // FLXDEBUG
    lout << "     FLX_DEBUG                     ";
    #if FLX_DEBUG
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLXDEBUG_COUT
    lout << "     FLX_DEBUG_COUT                ";
    #if FLX_DEBUG_COUT
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLX_BOOST_PO
    lout << "     FLX_BOOST_PO                  ";
    #if FLX_BOOST_PO
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  
  flxeigen_logInfo(lout);
  flxlib_logInfo(lout);
    
  // Compilation
    lout << " Compilation:" << std::endl;
    // Operating System
      lout << "   OS of compilation:              ";
      #ifdef __unix__
        lout << "UNIX ";
        #define FLX_OS_DET
      #endif
      #ifdef _WIN64
        lout << "WIN64 ";
        #define FLX_OS_DET
      #else
        #ifdef __WIN32__
          lout << "WIN32 ";
          #define FLX_OS_DET
        #endif
      #endif
      #ifdef __WINDOWS__
        lout << "Windows ";
        #define FLX_OS_DET
      #endif
      #ifdef __linux__
        lout << "Linux ";
        #define FLX_OS_DET
      #endif
      #ifndef FLX_OS_DET
        lout << "OTHER";
      #endif
      lout << std::endl;
    // Time of compilation
      lout << "   Compiled on " << __DATE__ << " at " << __TIME__ << std::endl;
    lout << "   Numeric limits ..." << std::endl;
    // tdouble
      lout << "     tdouble                       " << std::endl;
      lout << "        digits:                    " << numeric_limits<tdouble>::digits10 << std::endl;
      lout << "        epsilon:                   " << numeric_limits<tdouble>::epsilon() << std::endl;
      lout << "        min:                       " << (numeric_limits<tdouble>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<tdouble>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<tdouble>::round_style << std::endl;
    // tfloat
      lout << "     tfloat                        " << std::endl;
      lout << "        digits:                    " << numeric_limits<tfloat>::digits10 << std::endl;
      lout << "        epsilon:                   " << numeric_limits<tfloat>::epsilon() << std::endl;
      lout << "        min:                       " << (numeric_limits<tfloat>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<tfloat>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<tfloat>::round_style << std::endl;
    // double
      lout << "     double                        " << std::endl;
      lout << "        digits:                    " << numeric_limits<double>::digits10 << std::endl;
      lout << "        epsilon:                   " << numeric_limits<double>::epsilon() << std::endl;
      lout << "        min:                       " << (numeric_limits<double>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<double>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<double>::round_style << std::endl;
    // long double
      lout << "     long double                   " << std::endl;
      lout << "        digits:                    " << numeric_limits<long double>::digits10 << std::endl;
      lout << "        epsilon:                   " << numeric_limits<long double>::epsilon() << std::endl;
      lout << "        min:                       " << (numeric_limits<long double>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<long double>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<long double>::round_style << std::endl;
    // float
      lout << "     float                         " << std::endl;
      lout << "        digits:                    " << numeric_limits<float>::digits10 << std::endl;
      lout << "        epsilon:                   " << numeric_limits<float>::epsilon() << std::endl;
      lout << "        min:                       " << (numeric_limits<float>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<float>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<float>::round_style << std::endl;
    // int
      lout << "     int                           " << std::endl;
      lout << "        digits:                    " << numeric_limits<int>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<int>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<int>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<int>::round_style << std::endl;
    // long
      lout << "     long                          " << std::endl;
      lout << "        digits:                    " << numeric_limits<long>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<long>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<long>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<long>::round_style << std::endl;
    // unsigned int
      lout << "     unsigned int                  " << std::endl;
      lout << "        digits:                    " << numeric_limits<unsigned int>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<unsigned int>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<unsigned int>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<unsigned int>::round_style << std::endl;
    // unsigned long
      lout << "     unsigned long                 " << std::endl;
      lout << "        digits:                    " << numeric_limits<unsigned long>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<unsigned long>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<unsigned long>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<unsigned long>::round_style << std::endl;
    // tuint
      lout << "     tuint                         " << std::endl;
      lout << "        digits:                    " << numeric_limits<tuint>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<tuint>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<tuint>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<tuint>::round_style << std::endl;
    // tnlong
      lout << "     tnlong                        " << std::endl;
      lout << "        digits:                    " << numeric_limits<tnlong>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<tnlong>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<tnlong>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<tnlong>::round_style << std::endl;
    // tulong
      lout << "     tulong                        " << std::endl;
      lout << "        digits:                    " << numeric_limits<tulong>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<tulong>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<tulong>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<tulong>::round_style << std::endl;
    // size_t
      lout << "     size_t                        " << std::endl;
      lout << "        digits:                    " << numeric_limits<size_t>::digits10 << std::endl;
      lout << "        min:                       " << (numeric_limits<size_t>::min)() << std::endl;
      lout << "        max:                       " << (numeric_limits<size_t>::max)() << std::endl;
      lout << "        round_style:               " << numeric_limits<size_t>::round_style << std::endl;
      
    if (pversion) return;
    lout << "Fesslix: started with the configuration ..." << std::endl;
    // gauss.maxnumb
      lout << "  gauss.maxnumb                 ";
      lout << GaussIntegration::GaussPointMaxArraySize;
      lout << std::endl;
    // legendre.numb
      lout << "  legendre.numb                 ";
      lout << GlobalVar.LegendrePolyH_init_numb;
      lout << std::endl;
    // TOL
      lout << "  TOL                           " << GlobalVar.TOL() << std::endl;
    // leak-check
      lout << "  leak-check                    ";
      lout << bool2string(GlobalVar.is_leak_check());
      lout << std::endl;
    // floating point conversion
      GlobalVar.Double2String_log();
  
}

void flx_create_log(const std::string& logFile, const bool logTrunc) {
  if (logFile != "{screen}") {
    std::ofstream* ofLog;
    if ( logTrunc ) {
      ofLog = new std::ofstream(logFile.c_str());
    } else {
      ofLog = new std::ofstream(logFile.c_str(), std::ios_base::out | std::ios_base::app); 
    }
    if ( ofLog == NULL || !ofLog->is_open() ) {
      std::ostringstream ssV;
      ssV << "File (" << logFile << ") could not be opened.";
      throw FlxException("create_log_1", ssV.str() );
    } else {
      if (fesslix_useScreen) {
        GlobalVar.set_slogcout(ofLog,GlobalVar.get_cout());
      } else {
        GlobalVar.set_slogcout(ofLog,ofLog);
        GlobalVar.set_stdcerr(ofLog);
      }
    }
  }
  print_FLX_GNU(GlobalVar.slogcout(4));
  flx_print_version(GlobalVar.slogcout(4), true);
  // print date and time
    GlobalVar.slog(4) << "Time: ";
    flx_print_current_time(GlobalVar.slog(4));
    GlobalVar.slog(4) << std::endl;
      
  GlobalVar.slog(4) << "Fesslix: loglevel is " << GlobalVar.logLevel << std::endl;
  
  if (logFile == "{screen}" && fesslix_useScreen==false) {
    throw FlxException("flx_create_log_2","Configuration options 'log.usescreen=false' and 'log.file=\"{screen}\" are not allowed simultaneously");
  }
}

void flx_print_current_time_help(tuint i, ostream& os) {
  if (i<10) {
    os << "0";
  } 
  os << i;
}

void flx_print_current_time(ostream& os)
{
  time_t Zeitstempel = time(0);
  tm *nun = localtime(&Zeitstempel);
  os << nun->tm_year+1900 << "-";
  flx_print_current_time_help(nun->tm_mon+1,os);
  os << "-";
  flx_print_current_time_help(nun->tm_mday,os);
  os << " - ";
  flx_print_current_time_help(nun->tm_hour,os);
  os << ':';
  flx_print_current_time_help(nun->tm_min,os);
}

void flx_delete_FLXcout() {
  if (flx_engine_init_count>0) return;
  if (!fesslix_logOutput || flxcout==NULL) return;
  flxStreamAlloc* fsa = dynamic_cast<flxStreamAlloc*>(flxcout);
  #if FLX_DEBUG
    if (fsa==NULL) throw FlxException_Crude("delete_FLXcout");
  #endif
  delete fsa;
}

string flx_get_version()
{
  std::ostringstream ssV;
  ssV << "trunk." << VERSION_REV << " (" << VERSION_DATE << ")";
  return ssV.str();
}

void flx_print_version(std::ostream& so, bool pv)
{
  if (pv) {
    so << "Fesslix - Version";
  } else {
    so << "Fesslix";
  }
  so << ": " << flx_get_version() << std::endl;
  if (!pv) {
    flx_logInfo(so,true);
  } else {
    so << std::endl;
  }
}

void flxengine_load()
{
  // check for multiple runs
    static int Cruns = 0;
    ++Cruns;
    if (Cruns > 1) {
      std::ostringstream ssV;
      ssV << "flxengine_load can only be run once ...";
      throw FlxException("flxengine_load", ssV.str() );
    }
  FlxEngine = new FesslixMain;
}

void flxengine_unload()
{
  if (flx_engine_init_count==0) return;
  flx_engine_init_count--;
  if (flx_engine_init_count>0) return;
  if (FlxEngine) {
    delete FlxEngine;
    FlxEngine = NULL;
  }
}

const int flxengine_init(const bool cmd_parse, int argc, char** argv)
{
  flx_engine_init_count++;
  if (flx_engine_init_count>1) return RETURN_SUCCESS;
  try {
    setEXEdir();
    bool do_logInfo=DEFAULT_LOG_DO_LOGINFO;
    // set default values (in case BOOST::program_options is not used)
      tuint fpcPrec=DEFAULT_FPC_PREC, fpcType=DEFAULT_FPC_TYPE;
      tdouble fpcBvalU=DEFAULT_FPC_BVALU, fpcBvalL=DEFAULT_FPC_BVALL, tolV=DEFAULT_TOL;
      bool fpcDel0=DEFAULT_FPC_DEL0, fpcDelP=DEFAULT_FPC_DELP, logInput=DEFAULT_LOG_INPUT, logTrunc=DEFAULT_LOG_TRUNC;
      std::string gaussFile=DEFAULT_GAUSS_FILE, logFile=DEFAULT_LOG_FILE;
    bool cmd1st = DEFAULT_CMD1st;
      if (cmd_parse==false) logFile = "{screen}";
      
    #if FLX_BOOST_PO
    // Define Options  --- START --------------------------------------------------------------------------

      po::options_description generic(" Generic options");
      generic.add_options()
        ("help,h", "Show help about options")    
        ("author", "Show author information")
        ("version,v", "Show version information")
  //         ("license", "Show license information")
        ;

      std::string configFile = "{default}";
      // only allowed on the command line
      po::options_description arguments(" Arguments");
      arguments.add_options()
        ("file", po::value< std::vector<std::string> >(), "Parameter file to load")
        ("interactive,i", "Start Fesslix in interactive input mode")
        ("config-file", po::value<std::string>(&configFile)->default_value(configFile), 
            "Set a specific config-file")
        ;

      // Allowed on the command line as well as in the configuration file
      po::options_description config(" Configuration");
      config.add_options()
        ("flx.prgbar", po::value<bool>(&GlobalVar.prgBar)->default_value(GlobalVar.prgBar), 
            "Activate progress-bar feature for timeconsuming tasks.")
        ("flx.usescreen", po::value<bool>(&fesslix_useScreen)->default_value(fesslix_useScreen), 
            "false: output to 'cout' is redirected to the log-file.")
        ("fpc.prec", po::value<tuint>(&fpcPrec)->default_value(DEFAULT_FPC_PREC), 
            "Set the precision of floating point conversions.")
        ("fpc.type", po::value<tuint>(&fpcType)->default_value(DEFAULT_FPC_TYPE), 
            "Set the type of floating point conversions.")
        ("fpc.bvalu", po::value<tdouble>(&fpcBvalU)->default_value(DEFAULT_FPC_BVALU), 
            "Set the upper bound (type 2) of floating point conversions.")
        ("fpc.bvall", po::value<tdouble>(&fpcBvalL)->default_value(DEFAULT_FPC_BVALL), 
            "Set the lower bound (type 2) of floating point conversions.")
        ("fpc.del0", po::value<bool>(&fpcDel0)->default_value(DEFAULT_FPC_DEL0), 
            "Set removing of zeros in case of floating point conversions.")
        ("fpc.delp", po::value<bool>(&fpcDelP)->default_value(DEFAULT_FPC_DELP), 
            "Set removing of the leading plus in case of floating point conversions.")
        ("gauss.file", po::value<std::string>(&gaussFile)->default_value(DEFAULT_GAUSS_FILE), 
            "Set a specific gauss-file to load")
        ("gauss.numb", po::value<tuint>(&GaussIntegration::GaussPointArraySize)->default_value(DEFAULT_GAUSS_NUMB), 
            "Number of Gauss points to input - from 'gauss.file'.")
        ("gauss.maxnumb", po::value<tuint>(&GaussIntegration::GaussPointMaxArraySize)->default_value(DEFAULT_GAUSS_MAXNUMB), 
            "Maximal number of Gauss points to input - from 'gauss.file'.")
        ("legendre.numb", po::value<size_t>(&GlobalVar.LegendrePolyH_init_numb)->default_value(GlobalVar.LegendrePolyH_init_numb), 
            "Number of Legendre polynomials to initialize.")
        ("log.input", po::value<bool>(&logInput)->default_value(DEFAULT_LOG_INPUT), 
            "write input of Fesslix to the log-file")
        ("log.file", po::value<std::string>(&logFile)->default_value(logFile), 
            "Set a specific log-file")
        ("log.level", po::value<int>(&GlobalVar.logLevel)->default_value(GlobalVar.logLevel), 
            "Set the loglevel")
        ("log.options", po::value<bool>(&do_logInfo)->default_value(DEFAULT_LOG_DO_LOGINFO), 
            "log information about config-options")
        ("log.output", po::value<bool>(&fesslix_logOutput)->default_value(DEFAULT_LOG_OUTPUT), 
            "write output of Fesslix to the log-file")
        ("log.trunc", po::value<bool>(&logTrunc)->default_value(DEFAULT_LOG_TRUNC), 
            "overwrite existing log-files")
        ("MT19937.init.calls", po::value<tuint>(&GlobalVar.MT19937_init_calls)->default_value(GlobalVar.MT19937_init_calls), 
            "number of initial calls of MT19937 after initization (unsigned integer)")
        ("MT19937.init.rand", po::value<bool>(&GlobalVar.MT19937_init_RAND)->default_value(GlobalVar.MT19937_init_RAND), 
            "initialize MT19937 with RAND() instead of TIME")
        ("MT19937.init.seed", po::value<bool>(&GlobalVar.MT19937_init_seed)->default_value(GlobalVar.MT19937_init_seed), 
            "initialize MT19937 with seed value 'MT19937.init.seedvalue'")
        ("MT19937.init.seedvalue", po::value<tuint>(&GlobalVar.MT19937_init_seedvalue)->default_value(GlobalVar.MT19937_init_seedvalue), 
            "initial value in case 'MT19937.init.seed'=yes")
        ("flxvi.deactivate", po::value<bool>(&FlxVi_Deactivate)->default_value(false), 
            "deactivate the FlxVi-interface")
        ("TOL", po::value<tdouble>(&tolV)->default_value(DEFAULT_TOL),
            "A number with an absolute value smaller or equal than this threshold is considered to be zero.")
        ("pause", po::value<bool>(&fesslix_pauseAtEnd)->default_value(DEFAULT_PAUSE_AT_END), 
            "Wait for a keypress after execution.")
        ("cmd1st", po::value<bool>(&cmd1st)->default_value(cmd1st), 
            "Execute CMD first (and not last).")
        
        
  //         ("include-path,I", 
  //             po::value< vector<string> >()->composing(), 
  //             "include path")
        ;

      // Hidden options, will be allowed both on command line and
      // in config file, but will not be shown to the user.
      po::options_description hiddenCMD("Hidden options for command line");
      hiddenCMD.add_options()
          ("leak-check", "Run Fesslix, aware that leak-checking is done")   
          ;   
          
      po::options_description cmdline_options;
      cmdline_options.add(generic).add(arguments).add(config).add(hiddenCMD);

      po::options_description config_file_options;
      config_file_options.add(config);

      po::options_description visible("Allowed options in Fesslix");
      visible.add(generic).add(arguments).add(config);

      po::positional_options_description p;
      p.add("file", -1);

    // Define Options  --- END --------------------------------------------------------------------------
    #endif

    #if FLX_BOOST_PO
      // Parse Command Line Options  --- START --------------------------------------------------------------------------  
        
        int pos;
        flx_arg_CMD(argc, argv, pos, fesslix_CMDstr);
        
        po::variables_map vm;
        if (cmd_parse) {
          po::store(po::command_line_parser(pos, argv).
            options(cmdline_options).positional(p).run(), vm);
          po::notify(vm);
        }
        
      // Parse Command Line Options  --- END --------------------------------------------------------------------------

      // Parse Configuration File  --- START --------------------------------------------------------------------------  
        
        GlobalVar.pwd = "";
        get_pwd(GlobalVar.pwd);
        
        GlobalVar.configDir = "";
        get_config_dir_global(GlobalVar.configDir);
        
        // parse a user-defined configuration file
          std::ifstream configStream;
          if (configFile != "{default}") {
            configStream.open(configFile.c_str());
            if ( configStream ) {
              po::store(po::parse_config_file(configStream, config_file_options), vm);
              po::notify(vm);
              configStream.close();
            } else {
              GlobalVar.slog(1) << std::endl << "ERROR: specified config-file: File (" << configFile << ") could not be opened!" << std::endl << std::endl;
            }
          }
        
        // the very local one in $PWD
          if ( GlobalVar.configDir != "" ) {
            std::string configFileL = "fesslix.ini";
            std::ifstream configStream(configFileL.c_str());
            if ( configStream ) {
              po::store(po::parse_config_file(configStream, config_file_options), vm);
              po::notify(vm);
              configStream.close();
            }
          }
          
        // the local one
          std::string configDirL = "";
          if (!vm.count("leak-check")) {
            if (!get_config_dir_local(configDirL)) {
              configDirL = "";
            }
          } else {
            GlobalVar.set_leak_check_mode();
          }
          GlobalVar.configDir_local = configDirL;
          
          if (configDirL != GlobalVar.configDir && configDirL != "") {
            std::string configFileL = configDirL;
            configFileL.append("fesslix.ini");
            configStream.open(configFileL.c_str());
            if ( configStream ) {
              po::store(po::parse_config_file(configStream, config_file_options), vm);
              po::notify(vm);
              configStream.close();
            }
          }

        // the global one
          std::string configFileG = GlobalVar.configDir;
          configFileG.append("fesslix.ini");
          configStream.open(configFileG.c_str());
          if ( configStream ) {
            po::store(po::parse_config_file(configStream, config_file_options), vm);
            po::notify(vm);
            configStream.close();
          } else {
            GlobalVar.slog(1) << std::endl << "ERROR: global config-file: File (" << configFileG << ") could not be opened!" << std::endl << std::endl;
          }

      // Parse Configuration File  --- END --------------------------------------------------------------------------
    #else
      fesslix_logOutput=DEFAULT_LOG_OUTPUT;
      GlobalVar.configDir = "";
      get_config_dir_global(GlobalVar.configDir);
    #endif

    // set some global options
      GlobalVar.Double2String_setPrec(fpcPrec);
      GlobalVar.Double2String_setType(fpcType);
      GlobalVar.Double2String_setBValU(fpcBvalU);
      GlobalVar.Double2String_setBValL(fpcBvalL);
      GlobalVar.Double2String_setDel0(fpcDel0);
      GlobalVar.Double2String_setDelP(fpcDelP);      
      GlobalVar.set_TOL(tolV);
      GlobalVar.prelog_activated(logInput);
      GlobalVar.logLevel_strong_set(GlobalVar.logLevel);
      GlobalVar.cmd1st = cmd1st;
          
    #if FLX_BOOST_PO
      if (vm.count("help")) {
          flx_print_help(visible);
          if (fesslix_pauseAtEnd) { flx_take_a_break(); }
          return RETURN_SUCCESS;
      }
      if (vm.count("version")) {
          flx_print_version(*(GlobalVar.get_cout()));
          if (fesslix_pauseAtEnd) { flx_take_a_break(); }
          return RETURN_SUCCESS;
      }
      if (vm.count("author")) {
          flx_print_author();
          if (fesslix_pauseAtEnd) { flx_take_a_break(); }
          return RETURN_SUCCESS;
      }
    #endif

    // check arguments
      if ( cmd_parse && argc < 2 )  {
        *(GlobalVar.get_cout()) << "Please specify some options !!!" << std::endl << std::endl;
        #if FLX_BOOST_PO
          flx_print_help(visible);
        #else
          flx_print_help();
        #endif
        if (fesslix_pauseAtEnd) { flx_take_a_break(); }
        return RETURN_ERROR;
      }
      
    // Deal with the log-files
      flx_create_log(logFile, logTrunc);
    // create output-logFile
      
      if (!GlobalVar.check_logNOTcout()) fesslix_logOutput=false;
      if (fesslix_logOutput) {
        flxcout = new flxStreamAlloc(GlobalVar.get_cout(), GlobalVar.get_log());
      } else {
        flxcout = GlobalVar.get_cout();
      }
      
    // determin if interactive mode is to be called
      #if FLX_BOOST_PO
        fesslix_call_interact = (vm.count("interactive"));
      #endif
        
    // produce some logging-infos
      if (do_logInfo) {
        std::ostream& lout = GlobalVar.slog(4);
        flx_logInfo(lout,false);
        // Log-Options
          lout << "  log-options:" << std::endl;
          lout << "        level:                  " << GlobalVar.logLevel << std::endl;
          lout << "        log output    (output): " << bool2string(fesslix_logOutput) << std::endl;
          lout << "        log the input  (input): " << bool2string(logInput) << std::endl;
          lout << "        log-file        (file): " << logFile << std::endl;
          lout << "        truncate log-file:      " << bool2string(logTrunc) << std::endl;
          lout << "        log the config-options: " << bool2string(do_logInfo) << std::endl;
        // interactive
          lout << "  interactive:                  " << bool2string(fesslix_call_interact) << std::endl;
          if (fesslix_call_interact) {
          lout << "    flxvi.deactivate:           " << bool2string(FlxVi_Deactivate) << std::endl;
          }
        // pause
          lout << "  pause                         " << bool2string(fesslix_pauseAtEnd) << std::endl;
      }
      
    // create main object
      flxengine_load();
      FlxEngine->initialize(flxcout, GlobalVar.get_cerr(), gaussFile);
      
      #if FLX_DEBUG
      // testing
        FlxEngine->test();
      #endif

    #if FLX_BOOST_PO
      if (vm.count("file"))
      {
        for (tuint i = 0; i < vm["file"].as<std::vector<std::string> >().size(); ++i) {
          fesslix_fileVec.push_back(vm["file"].as<std::vector<std::string> >().operator[](i));
        }
      }
    #else
      for (int i=1;i<argc;++i) {
        fesslix_fileVec.push_back(argv[i]);
      }
    #endif
  } 
  catch ( FlxException &e) {
    FLXMSG("flxengine_init_101",1);
    *(GlobalVar.get_true_cerr()) << std::endl << e.what() << std::endl;
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << e.what() << std::endl;
      flx_print_RETURN_ERROR();
    }
    flxengine_unload();
    flx_delete_FLXcout();
    if (fesslix_pauseAtEnd) { flx_take_a_break(); }
    return RETURN_ERROR;
  }
  catch (std::exception const &e) {
    FLXMSG("flxengine_init_102",1);
    *(GlobalVar.get_true_cerr()) << std::endl << "ERROR: Whoops, something went wrong: " << e.what() << std::endl;
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << "ERROR: Whoops, something went wrong: " << e.what() << std::endl;
      flx_print_RETURN_ERROR();
    }
    flxengine_unload();
    flx_delete_FLXcout();
    if (fesslix_pauseAtEnd) { flx_take_a_break(); }
    return RETURN_ERROR;
  }
  catch(...) { 
    FLXMSG("flxengine_init_103",1);
    *(GlobalVar.get_true_cerr()) << std::endl << "ERROR: whoops!" << std::endl; 
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << "ERROR: whoops!" << std::endl; 
      flx_print_RETURN_ERROR();
    }
    flxengine_unload();
    flx_delete_FLXcout();
    if (fesslix_pauseAtEnd) { flx_take_a_break(); }
    return RETURN_ERROR;
  }
  return -1;
}

bool flx_arg_CMD(int argc, char** argv, int& pos, std::string& valueS )
{
  valueS = "";
  pos = 1;
  
  while (pos < argc) {
    std::string strV = argv[pos];
    if (strV == "--CMD") {
      for (int i = pos+1; i < argc; ++i) {
        valueS.append(argv[i]);
        valueS+=" ";
      }
      return true;
    }
    pos++;
  }
  return false;
}

const int flx_parse_file(const string& Expr, const bool is_file, const bool catch_all, bool do_log, int tabWidth)
{
  ReadStream *reader = NULL;
  try {
    if (is_file) {
      GlobalVar.slog(3) << "Fesslix: start parsing parameter file '" << Expr << "'" << std::endl;
      reader = new ReadStream( Expr.c_str(), do_log, tabWidth );
    } else {
      GlobalVar.slog(3) << "Fesslix: start evaluating the command from the command line: " << Expr << std::endl;
      reader = new ReadStream(Expr, do_log, tabWidth);  
    }
    FlxEngine->evaluate(reader);
    delete reader; reader = NULL;
    GlobalVar.logLevel_strong_reset();
  } 
  catch ( FlxEndE const &e) {
    FLXMSG("flx_parse_file_1",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) if (!fesslix_call_interact) throw;
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "end: 'end' was called" << std::endl;
    if (fesslix_call_interact) return RETURN_END;
    return RETURN_SUCCESS;
  }
  catch ( FlxException_NeglectInInteractive &e) {
    FLXMSG("flx_parse_file_2",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) if (!fesslix_call_interact) throw;
    *(GlobalVar.get_true_cerr()) << std::endl << e.what() << std::endl;
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << e.what() << std::endl;
    }
    return RETURN_ERROR;
  }
  catch ( FlxException &e) {
    FLXMSG("flx_parse_file_3",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) throw;
    *(GlobalVar.get_true_cerr()) << std::endl << e.what() << std::endl;
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << e.what() << std::endl;
    }
    GlobalVar.logLevel_strong_reset();
    return RETURN_ERROR;
  }
  catch (std::exception const &e) {
    FLXMSG("flx_parse_file_4",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) throw;
    *(GlobalVar.get_true_cerr()) << std::endl << "ERROR: Whoops, something went wrong: " << e.what() << std::endl;
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << "ERROR: Whoops, something went wrong: " << e.what() << std::endl;
    }
    return RETURN_ERROR;
  }
  catch(...) { 
    FLXMSG("flx_parse_file_5",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) throw;
    *(GlobalVar.get_true_cerr()) << std::endl << "ERROR: whoops!" << std::endl; 
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << "ERROR: whoops!" << std::endl; 
    }
    return RETURN_ERROR;
  }
  return RETURN_SUCCESS;
}

FlxObjBase* flx_read_file(const string& Expr, const bool is_file, const bool catch_all, bool do_log, int tabWidth)
{
  ReadStream *reader = NULL;
  try {
    if (is_file) {
      GlobalVar.slog(3) << "Fesslix: start reading parameter file '" << Expr << "'" << std::endl;
      reader = new ReadStream( Expr.c_str(), do_log, tabWidth );
    } else {
      GlobalVar.slog(3) << "Fesslix: start reading the command from the command line: " << Expr << std::endl;
      reader = new ReadStream(Expr, do_log, tabWidth);  
    }
    FlxObjBase* ob = FlxEngine->read_objects(reader);
    delete reader; reader = NULL;
    GlobalVar.logLevel_strong_reset();
    return ob;
  } 
  catch ( FlxException &e) {
    FLXMSG("flx_read_file_3",1);
    if (reader) { delete reader; reader=NULL; }
    if (!catch_all) throw;
    *(GlobalVar.get_true_cerr()) << std::endl << e.what() << std::endl;
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << e.what() << std::endl;
    }
    GlobalVar.logLevel_strong_reset();
    return NULL;
  }
  return NULL;
}

void flx_print_RETURN_SUCCESS() {
  GlobalVar.slogcout(4) << std::endl 
            << "Fesslix: Calculation completed. Fesslix(RETURN_SUCCESS)" << std::endl
            << "-------------------------------------------------------" << std::endl << std::endl;
}

void flx_print_RETURN_ERROR() {
  flx_print_version(GlobalVar.slogcout(3), true);
  GlobalVar.slogcout(4) << "Current Time: ";
  flx_print_current_time(GlobalVar.slogcout(4));
  GlobalVar.slogcout(4) << std::endl << std::endl;
  GlobalVar.slogcout(4) << "Fesslix: Termination due to an error. Fesslix(RETURN_ERROR)" << std::endl
                    << "-----------------------------------------------------------" << std::endl << std::endl;
}



FlxObjInteractive::FlxObjInteractive(const bool dolog, const string& ostreamV, const bool useVi)
: FlxObjOutputBase(dolog, ostreamV), ob(NULL), Ireader(NULL), useVi(useVi), resetInputLog(false), flxvi_loaded(false)
{

}


void FlxObjInteractive::print_prompt() {
  *(GlobalVar.get_cout()) << GlobalVar.get_flxPrompt();
}

void FlxObjInteractive::load_handle()
{
  #ifdef __unix__
    if (handle) throw FlxException_Crude("FlxObjInteractive::load_handle_unix_0");
    std::string libpathname = "libflxvi.so";
    handle = dlopen(libpathname.c_str(), RTLD_LAZY);
    if (handle==NULL) {
      ostringstream ssV;
      ssV << "Cannot find/open library '" << libpathname << "' (" << libpathname << ")." << std::endl;
      ssV << dlerror(); dlerror();
      GlobalVar.alert.alert("FlxObjInteractive::load_handle_unix_1",ssV.str());
    } else {
      flxvi_loaded = true;
    }
    dlerror();        // reset errors
  #elif defined(__WINDOWS__)
    if (sdl_library) throw FlxException_Crude("FlxObjInteractive::load_handle_win_0");
    std::string libpathname = "flxvi.dll";
    sdl_library = LoadLibrary(libpathname.c_str());
    if (sdl_library==NULL) {
      //ostringstream ssV;
      //ssV << "Cannot find/open library '" << libpathname << "'.";
      //GlobalVar.alert.alert("FlxObjInteractive::load_handle_win_1", ssV.str() );
    } else {
      flxvi_loaded = true;
    }
  #endif
}

void FlxObjInteractive::task() {
  // make sure that not interactive instance is already running
    if (!GlobalVar.allowInteractiveMode) {
      std::ostringstream ssV_2;
      ssV_2 << "Interactive input mode is already active.";
      throw FlxException_NeglectInInteractive("FlxObjInteractive::task_1", ssV_2.str() );
    }
    GlobalVar.allowInteractiveMode = false;
    if (!GlobalVar.check_logNOTcout()) {
      if (GlobalVar.prelog_isActive()) {
        resetInputLog=true;
        GlobalVar.prelog_activated(false);
      }
    }
  #if FLX_DEBUG
    if (ob || Ireader) throw FlxException_Crude("FlxObjInteractive::task_2");
  #endif
  try {
    GlobalVar.slog(3) << "interactive: start of the interactive input mode"<< std::endl;
    flxvi();
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
  } catch ( FlxExitE &e) {
    FLXMSG("FlxObjInteractive::task_1",1);
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "exit: 'exit;' was called - interactive mode will be closed." << std::endl;
    do_resetInputLog();
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
    return;
  } catch ( FlxEndE &e) {
    FLXMSG("FlxObjInteractive::task_2",1);
    GlobalVar.logLevel_strong_reset();
    do_resetInputLog();
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
    throw;
  } catch ( FlxException &e) {
    FLXMSG("FlxObjInteractive::task_3",1);
    GlobalVar.logLevel_strong_reset();
    do_resetInputLog();
    throw;
  } catch ( ... ) {
    FLXMSG("FlxObjInteractive::task_4",1);
    GlobalVar.logLevel_strong_reset();
    do_resetInputLog();
    throw;
  };
  do_resetInputLog();
}

void FlxObjInteractive::flxvi()
{
  // if use of flxVi is deactivated
    if (!useVi) {
      cmdline();
      return;
    }
  // if interface could not be loaded
    load_handle();
    if (flxvi_loaded==false) {
      cmdline();
      return;
    }
  // remember current streams
    const ostreamp rem_cerr = GlobalVar.get_cerr();
    const ostreamp rem_cout = GlobalVar.get_cout();
  try {
    flxvi_init();
    flxvi_exit();
    // back-initialize streams
      GlobalVar.set_stdcerr(rem_cerr);
      GlobalVar.set_slogcout(GlobalVar.get_log(),rem_cout);
  } catch (...) {
    FLXMSG("FlxObjInteractive::flxvi",1);
    flxvi_exit();
    // back-initialize streams
      GlobalVar.set_stdcerr(rem_cerr);
      GlobalVar.set_slogcout(GlobalVar.get_log(),rem_cout);
    throw;
  }
}

void FlxObjInteractive::flxvi_init()
{
  #ifdef __unix__
    if (handle==NULL) throw FlxException_Crude("FlxObjInteractive::flxvi_init_unix_1");
    typedef void (*hndlsymb_t)(FesslixMain*);
    hndlsymb_t hndlsymb = (hndlsymb_t) dlsym(handle, "flxvi_init");
    const char *dlsym_error = dlerror();
    if (dlsym_error) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxvi_init': " << dlsym_error;
      dlclose(handle);
      throw FlxException("FlxObjInteractive::flxvi_init_unix_2", ssV.str());
    }
    hndlsymb(fesslix_ptr);
  #elif defined(__WINDOWS__)
    if (sdl_library==NULL) throw FlxException_Crude("FlxObjInteractive::flxvi_win_1");
    FARPROC initializer = GetProcAddress(sdl_library,"flxvi_init");
    if (initializer == NULL) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxvi_init'";
      throw FlxException_NeglectInInteractive("FlxObjInteractive::flxvi_win_2", ssV.str() );
    }
    typedef void (*hndlsymb_t)(FesslixMain*);
    GlobalVar.slog(3) << "         convert symbol ..." << std::endl;
    hndlsymb_t hndlsymb = (hndlsymb_t) initializer;
    hndlsymb(fesslix_ptr);
  #endif
}

void FlxObjInteractive::flxvi_exit()
{
  #ifdef __unix__
    if (handle==NULL) throw FlxException_Crude("FlxObjInteractive::flxvi_exit_unix_1");
    typedef void (*hndlsymb_t)();
    hndlsymb_t hndlsymb = (hndlsymb_t) dlsym(handle, "flxvi_close");
    const char *dlsym_error = dlerror();
    if (dlsym_error) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxvi_close': " << dlsym_error;
      dlclose(handle);
      throw FlxException("FlxObjInteractive::flxvi_exit_unix_2", ssV.str());
    }
    hndlsymb();
  #elif defined(__WINDOWS__)
    if (sdl_library==NULL) throw FlxException_Crude("FlxObjInteractive::flxvi_exit_win_1");
    FARPROC initializer = GetProcAddress(sdl_library,"flxvi_close");
    if (initializer == NULL) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxvi_close'";
      throw FlxException_NeglectInInteractive("FlxObjInteractive::flxvi_exit_win_2", ssV.str() );
    }
    typedef void (*hndlsymb_t)();
    GlobalVar.slog(3) << "         convert symbol ..." << std::endl;
    hndlsymb_t hndlsymb = (hndlsymb_t) initializer;
    hndlsymb();
  #endif
}

void FlxObjInteractive::cmdline()
{
  (*(GlobalVar.get_cout())) << std::endl;
  print_GNU(*(GlobalVar.get_cout())); // cout and NOT sout !!!
  for (;;) {
    int wert;
    bool b_ml = false;
    std::ostringstream ssV;
    print_prompt();
    while ( (wert = std::cin.get()) != EOF && (wert != 10 || b_ml) ) {
      if ( wert == 92 && !b_ml) {
        b_ml = true;
      } else if ( wert == 10) {
        b_ml = false;
      } else {
        if (b_ml) {
          if ( (char)wert != ' ' && (char)wert != '\t' ) {
            ssV << '\\';
            b_ml = false;
          }
        }
        ssV << (char)wert;
      }
    }
    Ireader = new ReadStream(ssV.str(),true);
    data->ReadManager.push(Ireader);
    try {
      while ( Ireader->getNextType() != ReadStream::ENDOFFILE ) {
        // print the prompt for output of input to log
          GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
        ob = NULL;
        ob = EvaluateCmd->evaluateCmd();
        ob->exec();
        GlobalVar.logLevel_strong_reset();
        GlobalVar.slog_flush();
        delete ob; ob=NULL;
      }
      data->ReadManager.pop();
      delete Ireader;
    } catch ( FlxException_NeglectInInteractive &e) {
      FLXMSG("FlxObjInteractive::cmdline_1",1);
      *(GlobalVar.get_true_cerr()) << std::endl << e.what() << std::endl;
      GlobalVar.logLevel_strong_reset();
      if (GlobalVar.check_logNOTcout()) {
        GlobalVar.slog(1) << std::endl << e.what() << std::endl;
      }
      if (ob) { delete ob; ob=NULL; }
      data->ReadManager.pop();
      delete Ireader;
    } catch ( FlxException &e) {
      FLXMSG("FlxObjInteractive::cmdline_2",1);
      if (ob) { delete ob; ob=NULL; }
      data->ReadManager.pop();
      delete Ireader;
      throw;
    } catch (...) {
      FLXMSG("FlxObjInteractive::cmdline_3",1);
      if (ob) { delete ob; ob=NULL; }
      data->ReadManager.pop();
      delete Ireader;
      throw;
    }      
  }
}

void FlxObjInteractive::do_resetInputLog()
{
  #if FLX_DEBUG
    if (GlobalVar.allowInteractiveMode) {
      throw FlxException_Crude("FlxObjInteractive::do_resetInputLog");
    }
  #endif
  GlobalVar.allowInteractiveMode = true;
  if (!resetInputLog) return;
  resetInputLog = false;
  GlobalVar.prelog_activated(true);
}



FlxObjReadInteractive::FlxObjReadInteractive()
{
  // novi
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"interactive::novi"));
    ParaBox.insert("novi", "interactive::novi" );
}

FlxObjBase* FlxObjReadInteractive::read()
{
  read_optionalPara(false);
  return new FlxObjInteractive( get_doLog(), get_stream(), !get_optPara_bool("novi") );
}


