/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_fesslix_H
#define fesslix_fesslix_H

#include "flxobjcommon.h"
#include "flxlibs.h"
#include "flxostools.h"

#ifdef fesslix_fesslix_CPP
#else
  FLXKERNEL_EXPORT extern bool fesslix_pauseAtEnd;
  FLXKERNEL_EXPORT extern std::string fesslix_CMDstr;
  FLXKERNEL_EXPORT extern std::vector<std::string> fesslix_fileVec;
  FLXKERNEL_EXPORT extern bool fesslix_call_interact;
  FLXKERNEL_EXPORT extern bool FlxVi_Deactivate;
#endif
  
  
/**
* @brief The Main-Class
*/
class FLXKERNEL_EXPORT FesslixMain : public FlxReaderBase {
  public:
    FlxAddOnManager AddOnManager;
    FlxData dataBox;
    FlxFunctionReader *funReader;
    FlxEvaluateCmd *EvaluateCmd;
    FlxDefParaBox AllDefParaBox;
  private:
    static int Cinst;
    
    void load_FlxReaders(FlxCreateObjReaders* cor);
  public:
    FesslixMain ();
    ~FesslixMain ();
    
    /**
    * @brief Initialize some streams in Fesslix
    * @param coutV (cout)
    * @param cerrV (cerr)
    * @param gaussI (the input file stream to read in the gauss points - will be autom. deleted after input!)
    */
    void initialize(const ostreamp coutV, const ostreamp cerrV, const std::string& gaussFile);
    
    /**
    * @brief Evaluate the File (this will take some time - depending on the size of the file).
    * @Note this is a top-level command; don't use this recursively!
    */
    void evaluate ();
    void evaluate (ReadStream *readerV); 
    /**
    * @brief Read - but not evaluate - the File
    * @Note this is a top-level command; don't use this recursively!
    */
    FlxObjBase* read_objects (ReadStream *readerV);
    
    /**
    * @brief Evaluate an expression - when 'evaluate' is already active at a higher level
    */
    void evaluate_expr(const std::string exprStr);
    
    #if FLX_DEBUG
      void test ();
    #endif
};
#ifdef fesslix_fesslix_CPP
  FLXKERNEL_EXPORT FesslixMain* FlxEngine = NULL;
#else
  FLXKERNEL_EXPORT extern FesslixMain* FlxEngine;
#endif



  
FLXKERNEL_EXPORT void flx_print_author();
/**
* @brief logs some informations
*/
FLXKERNEL_EXPORT void flx_logInfo (std::ostream& lout, bool pversion);
FLXKERNEL_EXPORT std::string flx_get_version();
FLXKERNEL_EXPORT void flx_print_version(std::ostream& so, bool pv=false);
FLXKERNEL_EXPORT void flx_print_current_time(std::ostream& os);

/**
* @brief initialize Fesslix (read config-files,...)
* @returns -1 if end of function was reached
*/
FLXKERNEL_EXPORT const int flxengine_init(const bool cmd_parse=false, int argc=0, char** argv=NULL);
/**
* @brief deallocate the FesslixEngine
*/
FLXKERNEL_EXPORT void flxengine_unload();
/**
* @brief must be executed after flxengine_unload
*/
FLXKERNEL_EXPORT void flx_delete_FLXcout();

/**
* @brief evaluates the parameter file FileName
* @param is_file: true, if 'Expr' is FileName; false, if 'Expr' is command
* @param catch_all: true, if all errors should be catched
* @return only relevant if catch_all=true
*           if RETURN_RETURN_SUCCESS, the program can continue
*         otherwise, the execution should be stopped
*/
FLXKERNEL_EXPORT const int flx_parse_file(const std::string& Expr, const bool is_file, const bool catch_all, bool do_log=false, int tabWidth=8 );

/**
* @brief same as 'flx_parse_file', however, the file is only read but not executed
*/
FLXKERNEL_EXPORT FlxObjBase* flx_read_file(const std::string& Expr, const bool is_file, const bool catch_all, bool do_log=false, int tabWidth=8 );

FLXKERNEL_EXPORT void flx_print_RETURN_SUCCESS();
FLXKERNEL_EXPORT void flx_print_RETURN_ERROR();



  
  
  
/**
* @brief object class: start interactive mode
*/
class FLXKERNEL_EXPORT FlxObjInteractive : public FlxObjOutputBase, public FlxEvaluateCmdBase {
  private:
    // help-variables for cmdline-mode
      FlxObjBase* ob;
      ReadStream *Ireader;
    const bool useVi;                // false: flxVi is not used
    bool resetInputLog;
    bool flxvi_loaded;
    
    static FesslixMain* fesslix_ptr;
    
    // help-functions for cmdline-mode
      void do_resetInputLog();
      void print_prompt();
    // help-functions for flxvi-mode
      void load_handle();
    void task();
    
    /**
    * @brief load the flxvi-interface 
    * if this is not possible cmdline() is executed
    */
    void flxvi();
    void flxvi_init();
    void flxvi_exit();
    /**
    * @brief start the simple input mode
    */
    void cmdline();
  public:
    FlxObjInteractive ( const bool dolog, const std::string& ostreamV, const bool useVi );
    
    static void set_fesslix_ptr(FesslixMain* fesslix_ptrV) { fesslix_ptr = fesslix_ptrV; }
};


/**
* @brief object read class: for FlxObjInteractive
*/
class FlxObjReadInteractive : public FlxObjReadOutputBase {
  public:
    FlxObjReadInteractive();
    FlxObjBase* read (); 
};



#endif // fesslix_fesslix_H
