/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxlibs.h"
#include "flxlibs_internal.h"

#define fesslix_flxlibs_CPP

// Headers of the Libraries to include
  #include "flxBayUp_obj.h"

  

  
using namespace std;

#ifdef __unix__
  std::map<std::string, void*> AddOnList;
#elif defined(__WINDOWS__)
  std::map<std::string, HMODULE> AddOnList;
#endif

FlxAddOnManager* FlxObjReadLoadLib::AddOnManager = NULL;
   
#ifdef __unix__
  void handle_close(void* handle, const std::string& lname) {
    #if FLX_DEBUG
      GlobalVar.slogcout(1) << "unloading Add-on '" << lname << "' ..." << std::endl;
    #endif
    typedef void (*hndlsymb_t)();
    hndlsymb_t hndlsymb = (hndlsymb_t) dlsym(handle, "flxlib_close");
    const char *dlsym_error = dlerror();
    if (dlsym_error) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxlib_close': " << dlsym_error;
      dlclose(handle);
      throw FlxException_NeglectInInteractive("handle_close_unix_1", ssV.str() );
    }
    hndlsymb();
  }
#elif defined(__WINDOWS__)
  void handle_close(HMODULE sdl_library, const std::string& lname) {
    #if FLX_DEBUG
      GlobalVar.slogcout(1) << "unloading Add-on '" << lname << "' ..." << std::endl;
    #endif
    FARPROC initializer = GetProcAddress(sdl_library,"flxlib_close");
    if (initializer == NULL) {
      ostringstream ssV;
      ssV << "Cannot load symbol 'flxlib_close'";
      throw FlxException_NeglectInInteractive("handle_close_win_1", ssV.str() );
    }
    typedef void (*hndlsymb_t)();
    hndlsymb_t hndlsymb = (hndlsymb_t) initializer;
    hndlsymb();
  }
#endif
  
void FlxAddOnManager::delete_AddOns()
{
  if (bayup) delete BayUpBox;
  bool unload_fem = false;
  #ifdef __unix__
    for (map<std::string, void*>::iterator pos = AddOnList.begin(); pos != AddOnList.end(); ++pos) {
      if (pos->first == "fem") {
        unload_fem = true;
      } else {
      // free memory of allocated read-objects
        void* handle = pos->second;
        handle_close(handle,pos->first);
      }
    }
    if (unload_fem) {
      map<std::string, void*>::iterator pos = AddOnList.find("fem");
      if ( pos != AddOnList.end() ) {
        handle_close(pos->second,"fem");
      } else {
        throw FlxException_Crude("FlxAddOnManager::delete_AddOns_unix_1");
      }
    }
  #elif defined(__WINDOWS__)
    for (map<std::string, HMODULE>::iterator pos = AddOnList.begin(); pos != AddOnList.end(); ++pos) {
      if (pos->first == "fem") {
        unload_fem = true;
      } else {
        // free memory of allocated read-objects
          HMODULE sdl_library = pos->second;
          handle_close(sdl_library,pos->first);
      }
    }
  #endif
}

FlxAddOnManager::~FlxAddOnManager()
{
  #ifdef __unix__
    for (map<std::string, void*>::iterator pos = AddOnList.begin(); pos != AddOnList.end(); ++pos) {
        void* handle = pos->second;
      // close it ...
        if (pos->first!="matlab") {        // quick fix for segmentation fault
          dlclose(handle);
        }
    }
  #elif defined(__WINDOWS__)
    for (map<std::string, HMODULE>::iterator pos = AddOnList.begin(); pos != AddOnList.end(); ++pos) {
        HMODULE sdl_library = pos->second;
      // close it ...
        FreeLibrary(sdl_library);
    }
  #endif
}


FlxObjReadLoadLib::FlxObjReadLoadLib(FlxObjectReadBox* objReadBoxV)
: objReadBox(objReadBoxV)
{ 
  data->IgnoreBox.set_iL_recur("loadlib"); 
}

FlxObjBase* FlxObjReadLoadLib::read() {
  char c1 = 34;
  reader->getChar(c1,false);
  const string lname = reader->getWord(true,false);
  reader->getChar(c1,false);
  string pname = "";
  if (reader->whatIsNextChar()=='(') {
    reader->getChar('(');
    pname = reader->getText(false,false);
    reader->getChar(')');
  }
  
  FlxCreateObjReaders* cor=NULL;
  if (lname=="rnd") {
    if (AddOnManager->rnd) {
      ostringstream ssV;
      ssV << "Library '" << lname << "' has already been loaded.";
      throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_2", ssV.str(), reader->getCurrentPos());
    } else {
      AddOnManager->rnd = true;
      cor = new FlxCreateObjReaders_RND();
      cor->createFunReaders( data );
      cor->createObjReaders( objReadBox );
      delete cor;
    }
  } else if (lname=="bayup") {
    if (AddOnManager->bayup) {
      ostringstream ssV;
      ssV << "Library '" << lname << "' has already been loaded.";
      throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_4", ssV.str(), reader->getCurrentPos());
    } else {
      AddOnManager->bayup = true;
      BayUpBox = new FlxBayUpBox();
      cor = new FlxCreateObjReaders_BU();
      cor->createFunReaders( data );
      cor->createObjReaders( objReadBox );
      delete cor;
    }
  } else {        // try to load it as an AddOn
    bool llsuc = false;
    #ifdef __unix__
      std::string libpathname = "libflxlib" + lname + ".so";
      if (pname.empty()==false) libpathname = pname;
      GlobalVar.slog(3) << "loadlib: open library '" << lname << "' ..." << std::endl;
      void* handle = dlopen(libpathname.c_str(), RTLD_LAZY | RTLD_LOCAL); 
          // using the flag RTLD_GLOBAL causes problems with the variable 'flxlibs_interface flx_iface' if multiple modules are loaded 
      if (handle==NULL) {
        ostringstream ssV;
        ssV << "Cannot find/open library '" << lname << "' (" << libpathname << ").";
        ostringstream ssV2;
        ssV2 << dlerror() << std::endl << reader->getCurrentPos(); dlerror();
        throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_unix_1", ssV.str(), ssV2.str());
      }
      dlerror();        // reset errors
      GlobalVar.slog(3) << "         load symbols ..." << std::endl;
      typedef void (*hndlsymb_t)(FlxObjectReadBox*, FlxData*);
      hndlsymb_t hndlsymb = (hndlsymb_t) dlsym(handle, "flxlib_init");
      const char *dlsym_error = dlerror();
      if (dlsym_error) {
        ostringstream ssV;
        ssV << "Cannot load symbol 'flxlib_init': " << dlsym_error;
        dlclose(handle);
        throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_unix_2", ssV.str(), reader->getCurrentPos());
      }
      // insert in list
        pair<string, void*> Element(lname, handle);
        if ( ! AddOnList.insert(Element).second ) {
          ostringstream ssV;
          ssV << "Error during inserting " << lname << " in AddOnList.";
          throw FlxException("FlxObjReadLoadLib::read_unix_3", ssV.str());
        }
      hndlsymb(objReadBox,data);
      llsuc = true;
    #elif defined(__WINDOWS__)
      std::string libpathname = "flxlib" + lname + ".dll";
      if (pname.empty()==false) libpathname = pname;
      GlobalVar.slog(3) << "loadlib: open library '" << lname << "' ..." << std::endl;
      HMODULE sdl_library = LoadLibrary(libpathname.c_str());
      if (sdl_library==NULL) {
        ostringstream ssV;
        ssV << "Cannot find/open library '" << lname << "' (" << libpathname << ").";
        throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_win_1", ssV.str() );
      }
      GlobalVar.slog(3) << "         load symbols ..." << std::endl;
      FARPROC initializer = GetProcAddress(sdl_library,"flxlib_init");
      if (initializer == NULL) {
        ostringstream ssV;
        ssV << "Cannot load symbol 'flxlib_init'";
        throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_win_2", ssV.str() );
      }
      typedef void (*hndlsymb_t)(FlxObjectReadBox*, FlxData*);
      GlobalVar.slog(3) << "         convert symbol ..." << std::endl;
      hndlsymb_t hndlsymb = (hndlsymb_t) initializer;
      // insert in list
        pair<string, HMODULE> Element(lname, sdl_library);
        if ( ! AddOnList.insert(Element).second ) {
          ostringstream ssV;
          ssV << "Error during inserting " << lname << " in AddOnList.";
          throw FlxException("FlxObjReadLoadLib::read_win_3", ssV.str());
        }
      hndlsymb(objReadBox,data);
      llsuc = true;
    #endif
    if (llsuc==false) {
      ostringstream ssV;
      ssV << "Library '" << lname << "' does not exist.";
      throw FlxException_NeglectInInteractive("FlxObjReadLoadLib::read_5", ssV.str(), reader->getCurrentPos());
    }
  }
  
  GlobalVar.slog(3) << "loadlib: library '" << lname << "' loaded." << std::endl;
  return new FlxObjDummy(); 
}

// =======================================================================================================================
// LibFunManager
// =======================================================================================================================


FlxLibFunManager LibFunManager;


// =======================================================================================================================


FlxLibFunManager::~FlxLibFunManager()
{
  #ifdef __unix__
    for (map<std::string, void*>::iterator pos = LibList.begin(); pos != LibList.end(); ++pos) {
        void* handle = pos->second;
      // close it ...
        dlclose(handle);
    }
  #elif defined(__WINDOWS__)
    for (map<std::string, HMODULE>::iterator pos = LibList.begin(); pos != LibList.end(); ++pos) {
        HMODULE sdl_library = pos->second;
      // close it ...
        FreeLibrary(sdl_library);
    }
  #endif
}

handle_type FlxLibFunManager::get_handle(const string& libpn)
{
  #ifdef __unix__
    std::map<std::string, handle_type>::iterator it;
  #elif defined(__WINDOWS__)
    std::map<std::string, HMODULE>::iterator it;
  #endif
  it = LibList.find(libpn);
  if (it!=LibList.end()) return it->second;
  return NULL;
}

void FlxLibFunManager::insert_handle(const string& libpn, handle_type handle)
{
  pair<string, handle_type> Element(libpn, handle);
  if ( ! LibList.insert(Element).second ) {
    ostringstream ssV;
    ssV << "Error during inserting " << libpn << " in LibFunManager.";
    throw FlxException("FlxLibFunManager::insert_handle", ssV.str());
  }
}


// =======================================================================================================================


FlxObjReadLoadLibFun::FlxObjReadLoadLibFun(FlxObjectReadBox* objReadBoxV)
: objReadBox(objReadBoxV)
{ 
  data->IgnoreBox.set_iL_recur("loadlibfun"); 
  // type
    AllDefParaBox->insert(new FlxOptionalParaStream("fun","loadlibfun::type"));
    ParaBox.insert("type", "loadlibfun::type" );
}

FlxObjBase* FlxObjReadLoadLibFun::read()
{
  // -------------------------------------
  // read input information
  // -------------------------------------
    reader->getWord("name",false);
    reader->getChar('=',false);
    const std::string fname = reader->getWord(true,false);
    reader->getChar(',');
    reader->getWord("iname",false);
    reader->getChar('=',false);
    const std::string iname = reader->getWord(false,false);
    reader->getChar(',');
    reader->getWord("lib",false);
    reader->getChar('=',false);
    FlxString fstr(false, false);
    const std::string libpn = fstr.eval(false);
  read_optionalPara(false);
  const std::string tstr = get_optPara_string("type",true);
  tuint tid = 0;
  if (tstr == "fun") {
    tid = 0;
  } else if (tstr == "proc") {
    tid = 1;
  } else if (tstr == "mtx") {
    tid = 2;
  } else {
    std::ostringstream ssV;
    ssV << "unknown type '" << tstr << "'";
    throw FlxException_NeglectInInteractive("FlxObjReadLoadLibFun::read_0", ssV.str() );
  }
    
  // -------------------------------------
  // try to load the interface
  // -------------------------------------
    // ****************************
    // load the library
    // ****************************
      handle_type handle = NULL;
      handle = LibFunManager.get_handle(libpn);
      if (handle == NULL) {        // library not yet loaded
        GlobalVar.slog(3) << "loadlibfun: open library '" << libpn << "' ..." << std::endl;
        #ifdef __WINDOWS__
          handle = LoadLibrary(libpn.c_str());
        #else
          handle = dlopen(libpn.c_str(), RTLD_LAZY | RTLD_LOCAL); 
        #endif
        if (handle==NULL) {
          ostringstream ssV;
          ssV << "Cannot find/open library '" << libpn << "'";
          #ifdef __WINDOWS__
            throw FlxException_NeglectInInteractive("FlxObjReadLoadLibFun::read_win_1", ssV.str() );
          #else
            ostringstream ssV2;
            ssV2 << dlerror() << std::endl << reader->getCurrentPos(); dlerror();
            throw FlxException_NeglectInInteractive("FlxObjReadLoadLibFun::read_unix_1", ssV.str(), ssV2.str() );
          #endif
        }
        #ifndef __WINDOWS__
          dlerror();        // reset errors
        #endif
        LibFunManager.insert_handle(libpn, handle);
      }
    // ****************************
    // load the requested symbol
    // ****************************
      GlobalVar.slog(3) << "         load symbols ..." << std::endl;
      #ifdef __WINDOWS__
        FARPROC initializer = GetProcAddress( handle, iname.c_str() );
        if (initializer == NULL) {
          ostringstream ssV;
          ssV << "Cannot load symbol '" << iname.c_str() << "'";
          throw FlxException_NeglectInInteractive("FlxObjReadLoadLibFun::read_win_2", ssV.str() );
        }
              void* hndlsymb_void = initializer;
      #else
        void* hndlsymb_void = dlsym( handle, iname.c_str() );
        const char *dlsym_error = dlerror();
        if (dlsym_error) {
          ostringstream ssV;
          ssV << "Cannot load symbol '" << iname << "': " << dlsym_error;
          dlclose(handle);
          throw FlxException_NeglectInInteractive("FlxObjReadLoadLibFun::read_unix_2", ssV.str(), reader->getCurrentPos());
        }
      #endif
      switch (tid) {
          case 0:
          {
            flxfhndlsymb_t hndlsymb = (flxfhndlsymb_t) hndlsymb_void;
            data->FunBox.insert(fname,new FunReadFunLoadLibFun(fname,hndlsymb));
            break;
          }
          case 1:
          {
            flxfhndlsymbstrproc_t hndlsymb = (flxfhndlsymbstrproc_t) hndlsymb_void;
            objReadBox->insert(fname,new FlxObjReadLoadLibProc(fname,hndlsymb) );
            break;
          }
          case 2:
          {
            flxfhndlsymbmtx_t hndlsymb = (flxfhndlsymbmtx_t) hndlsymb_void;
            objReadBox->insert(fname,new FlxObjReadLoadLibMtx(fname,hndlsymb) );
            break;
          }
          default:
            throw FlxException_Crude("FlxObjReadLoadLibFun::read_3");
      }
  GlobalVar.slog(3) << "loadlibfun: " << tstr << "::" << fname << " linked to '" << libpn << "::" << iname << "." << std::endl;
  return new FlxObjDummy(); 
}


const tdouble FunLoadLibFun::calc()
{
  const std::string mtxname = (*ParaList_MtxName->begin())->eval();
  FlxSMtx* mtx = mtxConsts->get(mtxname,true);
  const tdouble* vp = mtx->get_internalPtr(true);
  int N = mtx->get_Ncoeff();
  return fhandle(vp, &N);
}

void FlxObjLoadLibProc::task()
{
  const std::string istr = strV->eval(false);
  int N = istr.length();
  fhandle(istr.c_str(),&N);
}

FlxObjBase* FlxObjReadLoadLibProc::read()
{
  FlxString* strV = new FlxString(true,false);
  try {
    read_optionalPara(false);
    return new FlxObjLoadLibProc(get_doLog(), fname, fhandle, strV );
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadLoadLibProc::read",1);
    delete strV;
    throw;
  }
}


void FlxObjLoadLibMtx::task()
{
  const std::string v_name = v->eval();
  const std::string w_name = w->eval();
  tuint N=0, M=0;
  const tdouble* vp = data->ConstMtxBox.get_Vec(v_name,N);
  tdouble* wp = data->ConstMtxBox.get_Vec(w_name,M);
  int Ni = N;
  int Mi = M;
  int err;
  fhandle(vp,&Ni,wp,&Mi,&err);
  if (err!=0) {
    ostringstream ssV;
    ssV << "Call to external library returned an error (" << err << ").";
    throw FlxException("FlxObjLoadLibMtx::task", ssV.str() );
  }
}

FlxObjBase* FlxObjReadLoadLibMtx::read()
{
  FlxMtxConstFun* v = new FlxMtxConstFun(true);
  FlxMtxConstFun* w = NULL;
  try {
    reader->getChar('-',false,false);
    reader->getChar('>',false,true);
    w = new FlxMtxConstFun(false);
    read_optionalPara(false);
    return new FlxObjLoadLibMtx(get_doLog(), fname, fhandle, v, w );
  } catch (FlxException &e) {
    delete v;
    if (w) delete w;
    throw;
  }
}



